/*
 * Copyright (c) 2005 Ichiro Satoh
 * All Rights Reserved.
 * written by Ichiro Satoh
 *
 * This program was developed by Ichiro Satoh at National Institute 
 * of Informatics as a part of a mobile object system, titled 
 * "AgentSpace".  All the programs of the system is developed by 
 * Ichiro Satoh and all rights are reserved by Ichiro Satoh. 
 *
 * Permission to use, copy, modify and distribute this software and
 * its documentation is hereby granted, provided that both the
 * copyright notice and this permission notice appear in all copies of
 * the software, derivative works or modified versions, and any
 * portions thereof, and that both notices appear in supporting 
 * documentation.
 */

/**
 * @author Ichiro Satoh
 * Created on 2005/05/8
 */

package agentspace.context;

import java.io.IOException;
import java.net.MalformedURLException;
import java.net.URL;
import java.util.Hashtable;

import agentspace.AgentIdentifier;
import agentspace.AgentServer;
import agentspace.Future;
import agentspace.InvalidAgentException;
import agentspace.Message;
import agentspace.NoSuchAgentException;
import agentspace.NoSuchResourceException;
import agentspace.registry.AgentRegistry;
import agentspace.registry.NotBoundException;
import agentspace.runtime.AgentContainer;
import agentspace.runtime.AgentRuntimeManager;

public class AgentContextManager extends ContextManager {
	AgentContainer container;
	
	public AgentContextManager(AgentRuntimeManager manager, AgentRegistry registry, AgentContainer container) {
		super(manager, registry);
		this.container = container;
	}
	
	public AgentIdentifier getIdentifier() {
		return container.getIdentifier();
	}

	public void setName(String name) {
		container.setName(name);
	}

	public String getName() {
		return container.getName();
	}

	public String getHost() {
		return AgentServer.getLocalHost();
	}
	
	public int getPort() {
		return AgentServer.getDefaultPort();
	}
	
	public URL getHostURL() {
		URL url = null;
		try {
			String host = AgentServer.getLocalHost();
			int port = AgentServer.getDefaultPort();
			if (port < 0) {
				url = new URL("matp://"+host);				
			}
			else {
				url = new URL("matp://"+host+":"+port);
			}
		}
		catch (MalformedURLException ex) {
			ex.printStackTrace();
		}
		return url;
	}
	
	public byte[] getLocalContent(String name) throws NoSuchResourceException {
		byte[] data = null;
		Hashtable table = container.getResources();
		if (table != null) {
			data = (byte[])table.get(name);
		}
		return data;
	}

	public void move(AgentIdentifier aid, URL url) {
		try {
			move(container, url, false, false);
		}
		catch (Exception ex) {
			ex.printStackTrace();
		}
	}
	
	public void store(String name) {
		class StoreHandler implements Runnable {
			String name;
			public StoreHandler(String name) {
				this.name = name;
			}
			public void run() {
				try {
					/*
					 *  need to fix  (by I.Satoh)
					 */
					byte[] data = manager.store(container);
					java.io.FileOutputStream fos = new java.io.FileOutputStream(name);
					fos.write(data, 0, data.length);
					fos.close();
				}
				catch (Exception ex) {
					ex.printStackTrace();
				}
			}
		}
		StoreHandler handler = new StoreHandler(name);
		manager.getThreadPool().startThread(handler);
	}

	public void move(URL url) {
		try {
			move(container, url,  false, false);
		}
		catch (Exception ex) {
			ex.printStackTrace();
		}
	}

	public void terminate() {
		terminate(container);
	}
	
	public void duplicate() {
		class DuplicateHandler implements Runnable {
			AgentContainer ac;
			public DuplicateHandler(AgentContainer ac) {
				this.ac = ac;
			}
			public void run() {
				manager.duplicate(ac);
			}
		}
		DuplicateHandler handler = new DuplicateHandler(container);
		manager.getThreadPool().startThread(handler);
	}

	public 	AgentIdentifier fetch(URL url) throws NoSuchAgentException, IOException {
		try {
			return getIdentifier(manager.fetch(url));
		} catch (InvalidAgentException e) {
			e.printStackTrace();
		} catch (NotBoundException e) {
			e.printStackTrace();
		} catch (ClassNotFoundException e) {
			e.printStackTrace();
		}
		return null;
	}
		

	public AgentIdentifier getIdentifier(AgentContainer ac) throws NotBoundException {
		return registry.getIdentifier(ac);
	}
	
	public AgentIdentifier[] getAgents() {
		if (registry.size() == 0) {
			return null;
		}
		else {
			AgentContainer[] acs = registry.getContainers();
			if (acs != null) {
				AgentIdentifier[] aids = new AgentIdentifier[registry.size()];
				for (int i = 0 ; i < registry.size() ; i++ ) {
					aids[i] = acs[i].getIdentifier();
				}
				return aids;
			}
			else {
				return null;
			}
		}
	}

	
	public String getAgentName(AgentIdentifier aid) throws NoSuchAgentException { 
		AgentContainer ac = registry.getContainer(aid);
		return ac.getName();
	}
	
	public AgentIdentifier[] getAgents(String name) {
		AgentContainer[] acs;
		try {
			acs = registry.lookup(name);
			if (acs != null) {
				AgentIdentifier[] aids = new AgentIdentifier[registry.size()];
				for (int i = 0 ; i < registry.size() ; i++ ) {
					aids[i] = acs[i].getIdentifier();
				}
				return aids;
			}
			
		} 
		catch (NotBoundException ex) {}
		return null;
	}
	
	public String getName(AgentIdentifier aid) throws NoSuchAgentException {
		return registry.getContainer(aid).getName(); 		
	}
	
	public Object invoke(AgentIdentifier aid, Message msg)
	throws Exception {
		if (msg.getFuture() == null) {
			msg.setFuture(new Future());
		}
		if (aid == null) {
			throw new NoSuchAgentException();
		}
		if (msg == null) {
			throw new NoSuchMethodException();
		}
		Future future = manager.invoke(aid, msg);
		return future.getReply();
	}
}