/*
 * Copyright (c) 2005 Ichiro Satoh
 * All Rights Reserved.
 * written by Ichiro Satoh
 *
 * This program was developed by Ichiro Satoh at National Institute 
 * of Informatics as a part of a mobile object system, titled 
 * "AgentSpace".  All the programs of the system is developed by 
 * Ichiro Satoh and all rights are reserved by Ichiro Satoh. 
 *
 * Permission to use, copy, modify and distribute this software and
 * its documentation is hereby granted, provided that both the
 * copyright notice and this permission notice appear in all copies of
 * the software, derivative works or modified versions, and any
 * portions thereof, and that both notices appear in supporting 
 * documentation.
 */

/**
 * @author Ichiro Satoh
 * Created on 2005/05/7
 */
package agentspace.runtime.object;

import java.lang.reflect.Constructor;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.lang.reflect.Modifier;

public class MethodUtil {
	/** An empty class array */
	private static final Class[] emptyClasses= new Class[0];
	/** An empty object array */
	private static final Object[] emptyObjects = new Object[0];
	
	public static Object invoke(Object object, String methodName,
			Class[] parameterTypes, Object[] parameterObjects)  
	throws NoSuchMethodException, IllegalAccessException, InvocationTargetException {
		if (parameterTypes == null) {
			parameterTypes = emptyClasses;
		}        
		if (parameterObjects == null) {
			parameterObjects = emptyObjects;
		}  
		Method method = getMethod(object.getClass(), methodName, parameterTypes, false);
		return method.invoke(object, parameterObjects);
	}
	
	public static Object invoke(Object object, Method method, Object[] parameterObjects)
	throws IllegalAccessException, InvocationTargetException {
		return method.invoke(object, parameterObjects);
	}
	
	public static Method getMethod(Class targetClass, String methodName, 
			Class[] classes, boolean asStatic) 
	throws NoSuchMethodException {
        int length = 0;
        if (classes != null) {
            length = classes.length;
        }
        Method[] meths = targetClass.getMethods();
        for( int i = 0 ; i < meths.length ; i++ ) {
            if (meths[i].getName().equals(methodName) &&
                    meths[i].getParameterTypes().length == length &&
                    (!asStatic || (meths[i].getModifiers() & Modifier.STATIC) != 0)) {
                Class[] methodArgClasses = meths[i].getParameterTypes();
                boolean isMatch = true;
                for(int j = 0 ; isMatch && j < length ; j++ ) {
                    isMatch = isMatch && 
                    (classes[j] == null ||
                            methodArgClasses[j].equals(classes[j]) ||
                            methodArgClasses[j].isAssignableFrom(classes[j]));
                }
                if (isMatch) {
                    return meths[i];
                }
            }
        }
        throw new NoSuchMethodException(methodName + " in " + targetClass);
    }
    
    /**
     * <p>Returns new instance of <code>klazz</code> created using constructor
     * with types <code>parameterTypes</code> and actual arguments <code>args</code>.</p>
     *
     * <p>The signatures should be assignment compatible.</p>
     *
     * @param cls the class to be constructed.
     * @param parameterObjects actual argument array
     * @param parameterTypes parameter types array
     * @return new instance of <code>class</code>
     */
    public static Object newInstance(Class cls, Class[] parameterTypes, Object[] parameterObjects)  
        throws IllegalAccessException, InvocationTargetException, InstantiationException {
        if (parameterTypes == null) {
            parameterTypes = emptyClasses;
        }        
        if (parameterObjects == null) {
            parameterObjects = emptyObjects;
        }  
        Constructor cst = getConstructor(cls, parameterTypes);
        return cst.newInstance(parameterObjects);
    }

    protected static Constructor getConstructor(Class targetClass, Class[] classes) { 
        int length = 0;
        if (classes != null) {
            length = classes.length;
        }
        Constructor[] csts = targetClass.getConstructors();
        for( int i = 0 ; i < csts.length ; i++ ) {
            Class[] cstArgClasses = csts[i].getParameterTypes();
            boolean isMatch = true;
            for(int j = 0 ; isMatch && j < length ; j++ ) {
                isMatch = isMatch && 
                (classes[j] == null ||
                        cstArgClasses[j].equals(classes[j]) ||
                        cstArgClasses[j].isAssignableFrom(classes[j]));
            }
            if (isMatch) {
                return csts[i];
            }
        }
        return null;
    }
}
